;-----------------------------------------------------------------------------
;* Servomotor ansteuern
;* Version: 0.2 	letzte Aenderung: 26.3.2010
;* Prozessor: ATtiny13A, interner Oszillator theoretisch 9.6MHz, gemessen 8HMz
;* Schaltschema: PB0 ueber Transistor zur Motorsteuerleitung
;*   ADC3 Poti als Eingabe zum austesten,
;*        oder Eingang vom Messgeraet: Negativer Puls startet Schuettelperiode
;* Autor: Rolf Pfister
;* Copyright: Freeware
;* History:
;* 24.3.2010	Erstellung
;* 26.3.2010	Auf negative Puls-Logik umgestellt (damit beim Ausstecken
;*              nicht dauernd geschuettelt wird)
;*
;* globale Registerbelegung:
;* r4 = ADC-Kanal
;* r25:r24 = Interruptzahler, 2000=20msec
;* r10...r15 Zeiten (20msec bis Tage)
;* r9 = PWM-Wert in 10usec-Schritten, fuer Servo Pulse von etwa 0.5 bis 2 ms
;.equ FREQTEST=1
.equ TEST1=1

.equ MINPWM=20
.equ MITTEPWM=150
.equ MAXPWM=250

;-----------------------------------------------------------------------------
.include "tn13def.inc"		; Definitionsdatei fuer den Prozessortyp
; Reset und Interruptvectoren	;VNr. Beschreibung
begin:	rjmp	main		; 1   Power On Reset
	reti			; 2   IRQ0 Handler
	reti			; 3   PCINT0
	rjmp	tc0over		; 4   Timer0 Overflow
	reti			; 5   EEPROM ready
	reti			; 6   Analog Comperator
	reti			; 7   Timer0 CompareA
	reti			; 8   Timer0 CompareB
	reti			; 9   Watchdog Interrupt
	rjmp	adcomplete	; 10  ADC Conversion
;-----------------------------------------------------------------------------
; Variablen im SRAM:
.equ SRAMSTART=0x60
.equ richtung=SRAMSTART		;Motoransteuerung (0b01 oder 0b10)
.equ sensor1adwertL=SRAMSTART+1
.equ sensor1adwertH=SRAMSTART+2 ;Wert von AD-Wandler fuer Stromsensor
.equ poti1adwertL=SRAMSTART+3
.equ poti1adwertH=SRAMSTART+4 ;Wert von AD-Wandler fuer Poti1
.equ poti2adwertL=SRAMSTART+5
.equ poti2adwertH=SRAMSTART+6 ;Wert von AD-Wandler fuer Poti2

;-----------------------------------------------------------------------------
; Start, Power On, Reset
main:	ldi r16, low(RAMEND)
        out SPL, r16  	     ; Init Stackpointer
	ldi r16, 0x80
	out CLKPR, r16		; Vorbereitung zum Prescaler setzen
	ldi r16, 0
	out CLKPR, r16		; Clk-Prescaler auf 1 setzen --> 9.6MHz
				; gemessen: 8HMz
	ldi r16, 0x03		;PB0, PB1 auf Ausgang
        out DDRB, r16		;Data Direction Register B setzen
	ldi r16, 0
	out PORTB, r16		;Motor aus
	ldi r17, 0x02
	out TIMSK0, r17		;Timer/Counter0 Overflow einschalten
	ldi r17, 1		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0B, r17		;Vorteiler auf 1 setzen
;Bei 9.6MHz wird somit in 0.104167us Schritten gezaehlt
;Um alle 10usec (96*0.104167) einen Interrupt zu erhalten muss der Zaehler
;jeweils bei 256-96=160 gestartet werden.
;bei 8MHz: 0.125us Schritte, 10/0.125=80, 256-80=176
;.equ	ZAEHLER=160		; bei 9.6MHz und ohne Vorteiler
.equ	ZAEHLER=176		; bei 8MHz und ohne Vorteiler
.equ	INTZAEHL=2000
	ldi r17, ZAEHLER
	out TCNT0, r17
	ldi r24, low(INTZAEHL)
	ldi r25, high(INTZAEHL)	; interruptzaehler fuer 20 msec setzen
	clr r10
	clr r11			; r11r10 = Millisec
	clr r12			; Sekunden
	clr r13			; Minuten
	clr r14			; Stunden
	clr r15			; Tage, auf 0 setzen
	sei			;Interrupt erlauben

; AD-Wandler initialisieren:
	ldi r16, 3
	mov r4, r16		;mit ADC3  beginnen
	ldi r16, 0		;Referenz=VCC
;	ldi r16, 0x40		;Referenz= etwa 1.1V
	add r16, r4
	out ADMUX, r16
;Prescaler: 3=8 4=16 ... 6=64, 9.6MHz/64=150kHz (muss im Bereich 50-200 liegen)
;AD-Wandlung braucht jeweils 13 Takte also etwa 87usec
.equ	PRESCALER=6
;r16 = Prescaler+0b11001000 (ADEN+ADSC+ADIE = Enable,Einzelstart,Interrupt)
;ADIF (0x10) wird automatisch vom AD-Wandler gesetzt wenn Wandlung fertig.
	ldi r16, 0b11001000 + PRESCALER
;test	out ADCSRA, r16		;Prescaler und Enable
		;ab hier wird also der AD-Wandler dauernd laufen

;PWM-Steuerung initialisieren:
	ldi r16, MAXPWM+1
	mov r9, r16

;Test: Frequenzmessung
.ifdef FREQTEST1
	cli
t1:	inc r16			;Schlaufe braucht 4 Takte, 8 pro Ganze Periode
	out PORTB, r16		;also 1.2MHz/8 = 150kHz an PB0
	rjmp t1			;oder 9.6MHz/8 = 1.2MHz
.endif
.ifdef FREQTEST2
	clr r17
	ldi r16, 20
t1:	inc r18
	out PORTB, r18
	rcall milliwait
	rjmp t1
.endif
.ifdef FREQTEST
t1:	rjmp t1
.endif
	
;-----------------------------------------------------------------------------
mainloop:

.ifdef TEST1
	ldi r18, MITTEPWM+1
	mov r9, r18
	ldi r16, 3
	rcall secwait		;3 sec warten
t1:	mov r9, r18
	ldi r16, 1
	rcall secwait		;langsam von mittlerer ...
	ldi r16, 1
	add r18, r16
	cpi r18, MAXPWM+2	;auf maximale Position fahren.
	brcs t1

	ldi r18, MITTEPWM+1
	mov r9, r18
	ldi r16, 3
	rcall secwait		;3 sec warten
	
t2:	mov r9, r18
	ldi r16, 1
	rcall secwait		;langsam von mittlerer ...
	ldi r16, 1
	sub r18, r16
	cpi r18, MINPWM+1	;auf minimale Position fahren.
	brcc t2
	rjmp mainloop

.else
	;; 50=0Grad, 200=180Grad --> xGrad=150/180*x+50
	ldi r16, 88		; 45Grad = Ausposition
	mov r9, r16		; Servo stellen
	lds r17, poti2adwertH
	lds r16, poti2adwertL	;r17:r16=adwert
	ldi r18, 3
	rcall div18		; auf Bereich 0...128 verkleinern
	cpi r16, 96		; Obere Poti-Stellung ?
	brcc minus		; ja-->
	cpi r16, 32		; Untere Poti-Stellung ?
	brcs plus		; ja-->
mitte:				;Poti ungefaehr in Mitte: automatische Sequenz
	mov r16, r12
	andi r16, 0x0F
	brne mainloop		; ca 16 Sec auf Aus-Position bleiben
	ldi r16, 162		; 135Grad
	mov r9, r16
	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait		; 0.5 Sec um neue Position zu erreichen
	ldi r16, 2		; 2 Sec auf dieser Position bleiben
	rcall secwait
	rjmp mainloop
plus:
	ldi r16, 162		; 135Grad
	mov r9, r16
	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait		; 0.5 Sec um neue Position zu erreichen
	ldi r16, 2		; 2 Sec auf dieser Position bleiben
	rcall secwait
	ldi r16, 88		; Servo auf Aus-Position
	mov r9, r16
	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait		; 0.5 Sec um Aus-Position zu erreichen
	rjmp mainloop
minus:
	ldi r16, 88		; Servo auf Aus-Position
	mov r9, r16
	rjmp mainloop
.endif

div18:		;r17:r16 /= 2^r18 ,r18 zerstoert
	asr r17
	ror r16
	dec r18
	brne div18
	ret

secwait:	; r16 Sekunden lang warten
	push r16
	push r17
	push r18
	mov r18, r16
	add r18, r16
_sw1:	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait
	dec r18
	brne _sw1
	pop r18
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; Interrupt-Routine
; Registerbelegungen:
; r25:r24 interruptzaehler
; r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
tc0over:push r16
	push r17
	in r17, SREG	;Statusregister sichern
	push r17
	ldi r17, ZAEHLER
	out TCNT0, r17
.ifdef FREQTEST
	out PORTB, r24
.endif
.ifndef FREQTEST
pwmsteuerung:			; r9=PWM-Wert in 10usec-Schritten
	tst r25
	brne pwmoff
	cp r24, r9
	brcc pwmoff
pwmon:	sbi PORTB, 0		; Clear Bit 0
	rjmp pwmend
pwmoff:	cbi PORTB, 0		; Set Bit 0
pwmend:
.endif
	sbiw r25:r24, 1
	breq millisekzaehlen
pop17reti:
	pop r17
	out SREG, r17	;Statusregister zurueckholen
	pop r17
	pop r16
	reti
millisekzaehlen:
	ldi r24, low(INTZAEHL)
	ldi r25, high(INTZAEHL)
	ldi r16, 20
	add r10, r16		;Millisekunden erhoehen
	brcc L1
	inc r11
L1:	ldi r16, low(1000)
	ldi r17, high(1000)	;r17:r16 = 1000
	cp r10, r16
	brne end1	;fertig wenn msec!=1000
	cp r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
	cp r13, r16
	brne end1	;fertig wenn min!=60
	clr r13		;Minuten auf 0 setzen
	inc r14		;Stunden erhoehen
	ldi r16, 24
	cp r14, r16
	brne end1	;fertig wenn std!=24
	clr r14		;Stunden auf 0 setzen
	inc r15		;Tage erhoehen
end1:	rjmp pop17reti

;-----------------------------------------------------------------------------
; Interrupt-Routine fuer AD-Wandler
; Rueckgabewert: sensor1adwert, poti1wert, poti2wert
adcomplete:
	push r16
	push r17
	in r17, SREG	;Statusregister sichern
	push r17
	push r18
	in r17, ADCL	;L muss zuerst gelesen werden!
	in r16, ADCH
	mov r18, r4
	cpi r18, 3
	breq _ad3
	cpi r18, 2
	breq _ad2
_ad1:	sts poti1adwertL, r17 	;Messwert im RAM
	sts poti1adwertH, r16	;zwischenspeichern.
	rjmp _ad4
_ad2:	sts sensor1adwertL, r17 	;Messwert im RAM
	sts sensor1adwertH, r16		;zwischenspeichern.
	rjmp _ad4
_ad3:	sts poti2adwertL, r17 	;Messwert im RAM
	sts poti2adwertH, r16	;zwischenspeichern.
_ad4:
;	dec r4		;automatisch naechster AD-Kanal
;	brne _ad5
;	ldi r16, 3
;	mov r4, r16		;ADC3 als Anfangswert
_ad5:	ldi r16, 0		;Referenz=VCC
	add r16, r4		;ADC-Kanal
	out ADMUX, r16
	sbi ADCSRA, ADSC	;Starte naechste AD-Wandlung
	pop r18
	rjmp pop17reti

;-----------------------------------------------------------------------------
milliwait:			;Wartezeit: r17:r16 msec, min 1 max 999
	; oder bei 20ms Schrittweite: min 20 max 980, durch 20 teilbar
	push r16
	push r17
	push r18
	push r19
	mov r18, r10
	mov r19, r11		;r19r18 = sollzeit = aktuelle Millisec
	add r18, r16
	adc r19, r17		;sollzeit += Wartezeit
	ldi r16, 0xE8		;auf 1000 testen
	cpi r19, 0x03		;oberes Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
	brne subtra		;sollzeit>1000 --> subtrahiere 1000
	cp r18, r16		;unteres Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
subtra:	sub r18, r16
	sbci r19, 0x03		;wenn sollzeit>=1000 dann sollzeit-=1000
loop1:	cp r18, r10
	brne loop1
	cp r19, r11		;auf sollzeit warten
	brne loop1
	pop r19
	pop r18
	pop r17
	pop r16
	ret
