;-----------------------------------------------------------------------------
;* Solarzellen-Nachfuehrung mit Modellbau-Servo
;* Version: 0.5 	letzte Aenderung: 11.4.2010
;* Prozessor: ATtiny13A, interner Oszillator 8MHz
;* Schaltschema: PB0 direkt zur Steuerleitung 
;*     ADC1, ADC3 Analogeingaenge von Phototransistoren,
;*     (PT1331C mit 10k am Collector, sollte bei Sonne weniger 0.1 V ergeben)
;* Autor: Rolf Pfister
;* Copyright: Freeware
;* History:
;* 11.4.2010	Erstellung aus solnach.asm und pwmservo.asm
;*
;* globale Registerbelegung:
;* r20 = ADC-Kanal
;* r0 = Sicherung Statusregister im AD-Wandler-Interrupt
;* r23:r22 = aktueller PWM-Wert im Timer
;* r25:r24 = PWM-Wert in 1usec-Schritten, (Servo Pulse von etwa 0.5 bis 2.5 ms)
;* r10...r15 Zeiten (msec bis Tage)
;* r9 = Zaehler im Timerinterrupt

;Werte in Microsekunden:
.equ MINPWM=600
.equ MITTEPWM=1500
.equ MAXPWM=2400

;.equ FREQTEST=1
;.equ FREQTEST2=1
;.equ TIMERTEST=1
;.equ TEST1=1

.equ PULSPB0=1
;.equ PULSPB3=1

;-----------------------------------------------------------------------------
.include "tn13def.inc"		; Definitionsdatei fuer den Prozessortyp
; Reset und Interruptvectoren	;VNr. Beschreibung
begin:	rjmp	main		; 1   Power On Reset
	reti			; 2   IRQ0 Handler
	reti			; 3   PCINT0
	rjmp	tc0over		; 4   Timer0 Overflow
	reti			; 5   EEPROM ready
	reti			; 6   Analog Comperator
	reti			; 7   Timer0 CompareA
	rjmp	tc0compb	; 8   Timer0 CompareB
	reti			; 9   Watchdog Interrupt
	rjmp	adcomplete	; 10  ADC Conversion
;-----------------------------------------------------------------------------
; Variablen im SRAM:
.equ SRAMSTART=0x60
.equ reserve=SRAMSTART		;
.equ adc1wertL=SRAMSTART+1
.equ adc1wertH=SRAMSTART+2	;Eingelesener Wert von ADC1
.equ adc2wertL=SRAMSTART+3
.equ adc2wertH=SRAMSTART+4	;Eingelesener Wert von ADC2
.equ adc3wertL=SRAMSTART+5
.equ adc3wertH=SRAMSTART+6	;Eingelesener Wert von ADC3

;-----------------------------------------------------------------------------
; Start, Power On, Reset
main:	ldi r16, low(RAMEND)
        out SPL, r16  	     ; Init Stackpointer
	ldi r16, 0x80
	out CLKPR, r16		; Vorbereitung zum Prescaler setzen
	ldi r16, 0
	out CLKPR, r16		; Clk-Prescaler auf 1 setzen --> 9.6MHz
	ldi r16, 0x51		; Wert ausprobieren fuer 8MHz
	out OSCCAL, r16		; Oszillator kallibrieren --> 8MHz

;; Init-Code
	ldi r16, 0x03		;PB0, PB1 auf Ausgang
.ifdef PULSPB3
	ldi r16, 0x0B		;PB0, PB1, PB3 auf Ausgang
.endif
        out DDRB, r16		;Data Direction Register B setzen
	ldi r16, 0
	out PORTB, r16		;Motor aus

.ifdef FREQTEST
T1:	inc r16			; Test fuer Frequenzmessung
	out PORTB, r16		; Schlaufe braucht 4 Takte, eine Periode also 8
	rjmp T1			; Bei 8MHz muesste an B0 1MHz zu messen sein.
				; gemessen: 1.003 MHz
.endif

;; Timer auf 1MHz (1usec) und 250 setzen: 250*1usec=0.25msec
;; bei jedem 80. Durchgang (alle 20msec) PWM neu starten
	ldi r17, 0x0A		;Interruptenable setzen: B=8 A=4 OVR=2
	out TIMSK0, r17
;	ldi r17, 2		;CTC-Modus
; Ohne benutzen von Pin OC0B:
	ldi r17, 3		;PWM-Modus 3 oder 7
; Wenn Pin OC0B (PB1 beim tiny13a) automatisch gesetzt werden soll:
;	ldi r17, (1<<COM0B1)+3	;OC0B benutzen und PWM-Modus 3 oder 7
	out TCCR0A, r17		;Timer-Modus setzen
; Fuer PWM-Modus 7:
	ldi r17, 2+8		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
; Fuer andere Moden:
;	ldi r17, 2		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0B, r17		;Vorteiler auf 8 setzen (8MHz/8=1MHz)
	ldi r17, 249
	ldi r17, 239		;test
	out OCR0A, r17		;TOP=249 fuer teilen durch 250
	ldi r17, 0
	out OCR0B, r17		;PWM-Wert loeschen
	ldi r17, 0
	out TCNT0, r17		;Counter auf Startwert setzen
	ldi r17, 1
	mov r9, r17		;Durchlaufzaehler setzen

	clr r10
	clr r11			; r11r10 = Millisec
	clr r12			; Sekunden
	clr r13			; Minuten
	clr r14			; Stunden
	clr r15			; Tage, auf 0 setzen

	sei			;Interrupt erlauben

; AD-Wandler initialisieren:
	ldi r20, 3		;mit ADCx beginnen
	ldi r16, 0		;Referenz=VCC
;	ldi r16, 0x40		;Referenz= etwa 1.1V
	add r16, r20
	out ADMUX, r16
;Prescaler: 3=8 4=16 ... 6=64, 8MHz/64=125kHz (muss im Bereich 50-200 liegen)
;AD-Wandlung braucht jeweils 13 Takte also etwa 104usec
.equ	PRESCALER=6
;r16 = Prescaler+0b11001000 (ADEN+ADSC+ADIE = Enable,Einzelstart,Interrupt)
;ADIF (0x10) wird automatisch vom AD-Wandler gesetzt wenn Wandlung fertig.
	ldi r16, 0b11001000 + PRESCALER
	out ADCSRA, r16		;Prescaler und Enable
		;ab hier wird also der AD-Wandler dauernd laufen

;PWM-Steuerung initialisieren:
	ldi r24, low(MITTEPWM)
	ldi r25, high(MITTEPWM) ;Mittelstellung

.ifdef FREQTEST2
T1:	rjmp T1
.endif

.ifdef TIMERTEST
T1:	inc r18
	out PORTB, r18
	ldi r16, 1
	rcall milliwait1	;gemessen: 485Hz 2.06ms, (mit 100: 208.2ms)
	rjmp T1
.endif

;-----------------------------------------------------------------------------
mainloop:

.ifdef TEST1
;Motortest: von der Mitte her auf die Grenzen fahren
	ldi r24, low(MITTEPWM)
	ldi r25, high(MITTEPWM)
	ldi r16, 3
	rcall secwait		;3 sec warten
t1:	ldi r16, 1
	rcall secwait		;langsam von mittlerer ...
	subi r24, low(-10)
	sbci r25, high(-10)
	ldi r16, low(MAXPWM)	;auf maximale Position fahren.
	ldi r17, high(MAXPWM)	;auf maximale Position fahren.
	cp r24, r16
	cpc r25, r17
	brcs t1
	ldi r16, 3
	rcall secwait		;3 sec auf Max-Position warten

	ldi r24, low(MITTEPWM)
	ldi r25, high(MITTEPWM)
	ldi r16, 3
	rcall secwait		;3 sec warten
	
t2:	ldi r16, 1
	rcall secwait		;langsam von mittlerer ...
	subi r24, 10
	sbci r25, 0
	ldi r16, low(MINPWM)	;auf minimale Position fahren.
	ldi r17, high(MINPWM)	;auf minimale Position fahren.
	cp r24, r16
	cpc r25, r17
	brcc t2
	ldi r16, 3
	rcall secwait		;3 sec auf Min-Position warten
	rjmp mainloop

.else
;endgueltige Variante
	ldi r16, 100
	rcall milliwait1
	lds r16, adc1wertL
	lds r17, adc1wertH	;unterer Lichtwert
	lds r18, adc2wertL
	lds r19, adc2wertH	;oberer Lichtwert
	sub r16, r18
	sbc r17, r19
	breq motstop
	brcs nachunten		;unterer Wert kleiner --> unten heller
nachoben:
	ldi r21, high(500)
	cpi r18, low(500)	;oberer Lichtwert
	cpc r19, r21		;auf minimale Helligkeit testen
	brcc motstop		;zu dunkel --> keine Nachfuehrung machen
	ldi r21, high(MAXPWM)
	cpi r24, low(MAXPWM)
	cpc r25, r21		;r25:r24 == MAXPWM ?
	breq motstop		;ja --> am Anschlag
	subi r24, low(-1)
	sbci r25, high(-1)	;r25:r24 += 1
	rjmp mainloop
nachunten:
	lds r16, adc1wertL
	lds r17, adc1wertH	;unterer Lichtwert
	ldi r21, high(500)
	cpi r16, low(500)
	cpc r17, r21		;auf minimale Helligkeit testen
	brcc motstop		;zu dunkel --> keine Nachfuehrung machen
	ldi r21, high(MINPWM)
	cpi r24, low(MINPWM)
	cpc r25, r21		;r25:r24 == MINPWM ?
	breq motstop		;ja --> am Anschlag
	subi r24, 1
	sbci r25, 0		;r25:r24 -= 1
	rjmp mainloop
motstop:		;Motor stoppen
	ldi r16, 1
	rcall secwait
	rjmp mainloop
.endif

.ifdef DIVISIONEN
div18:		;r17:r16 /= 2^r18 ,r18 zerstoert
	asr r17
	ror r16
	dec r18
	brne div18
	ret
.endif
	
secwait:	; r16 Sekunden lang warten
	push r16
	push r17
	push r18
	mov r18, r16
	add r18, r16
_sw1:	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait
	dec r18
	brne _sw1
	pop r18
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; OVR-Interrupt-Routine, wird alle 0.25msec aufgerufen
; Registerbelegungen:
; r9=Durchlaufzaehler, jeder 80.Durchlauf den PWM-Wert setzen (80*0.25=20msec)
; r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
tc0over:push r16
	in r16, SREG	;Statusregister sichern
	push r16
	dec r9			;Durchlaufzaehler -= 1
.ifdef FREQTEST2
	out PORTB, r9		;gemessen: 1.913 MHz
	rjmp pop16reti
.endif
	breq tc2		;wenn r9==0 --> Puls starten
	ldi r16, 1
	cp r9, r16		;r9==1 ?
	brne tc1a		;nein-->
	mov r22, r24		;ja: jeder 80. Durchlauf neuer PWM-Wert starten
	mov r23, r25		;r23:r22 = r25:r24
tc1a:	tst r23
	brne tc1b		; >250 -->
	cpi r22, 250		; r23:r22 >= 250 ?
	brcc tc1b		; ja -->
	out OCR0B, r22	;PWM-Wert zetzen, ist erst beim naechsten mal aktiv
	clr r22
tcend:	tst r9
	breq tc3
	ldi r16, 3		;ist r9 durch 4
	and r16, r9		;teilbar?
	breq tc4		;ja--> Millsekunden raufzaehlen
pop16reti:
	pop r16
	out SREG, r16	;Statusregister zurueckholen
	pop r16
	reti
tc1b:	subi r22, 250
	sbci r23, 0		; r23:r22 -= 250
	ldi r16, 250
	out OCR0B, r16	;PWM-Wert zetzen, ist erst beim naechsten mal aktiv
	rjmp tcend
tc2:
.ifdef PULSPB0
	sbi PORTB, 0		;Puls an PB0 starten
.endif
.ifdef PULSPB3
	sbi PORTB, 3		;Puls an PB3 starten
.endif
	rjmp tc1a
tc3:	ldi r16, 80
	mov r9, r16
tc4:	sei
millisekzaehlen:
	push r17
	inc r10		;Millisekunden erhoehen
	brne L1
	inc r11
L1:	ldi r16, low(1000)
	ldi r17, high(1000)	;r17:r16 = 1000
	cp r10, r16
	cpc r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
	cp r13, r16
	brne end1	;fertig wenn min!=60
	clr r13		;Minuten auf 0 setzen
	inc r14		;Stunden erhoehen
	ldi r16, 24
	cp r14, r16
	brne end1	;fertig wenn std!=24
	clr r14		;Stunden auf 0 setzen
	inc r15		;Tage erhoehen
end1:	pop r17
	rjmp pop16reti

tc0compb:			;Ende des PWM-Pulses
.ifdef PULSPB0
	cbi PORTB, 0
.endif
.ifdef PULSPB3
	cbi PORTB, 3
.endif
	reti

;-----------------------------------------------------------------------------
; Interrupt-Routine fuer AD-Wandler
; Rueckgabewert: adc1wert, adc2wert, adc3wert
adcomplete:
	push r16
	push r17
	in r17, ADCL	;L muss zuerst gelesen werden!
	in r16, ADCH
	sei		;Interrupt erlauben (fuer schnelles PWM)
	in r0, SREG	;Statusregister sichern
	cpi r20, 3
	breq _ad3
	cpi r20, 2
	breq _ad2
_ad1:	sts adc1wertL, r17 	;Messwert im RAM
	sts adc1wertH, r16	;zwischenspeichern.
	rjmp _ad4
_ad2:	sts adc2wertL, r17 	;Messwert im RAM
	sts adc2wertH, r16		;zwischenspeichern.
	rjmp _ad4
_ad3:	sts adc3wertL, r17 	;Messwert im RAM
	sts adc3wertH, r16	;zwischenspeichern.
_ad4:
	dec r20		;automatisch naechster AD-Kanal
	brne _ad5
	ldi r20, 3		;ADC3 als Anfangswert
_ad5:	ldi r16, 0		;Referenz=VCC
;	ldi r16, 0x40		;Referenz= etwa 1.1V
	add r16, r20		;ADC-Kanal
	out ADMUX, r16
	sbi ADCSRA, ADSC	;Starte naechste AD-Wandlung
	out SREG, r0	;Statusregister zurueckholen
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; milliwait  optimierte Variante  Wartezeit 1 bis 1000 Millisekunden
; in r11:r10 wird durch einen Timerinterrupt die aktuelle Millisekunde
; gespeichert (16Bit-Zahl von 0 bis 999).
; Die aktelle Sekunde ist in r12 (wird hier noch nicht benutzt, da nur
; bis maximal 1000ms gewartet werden soll).
milliwait1:			;Wartezeit: r16, 1 bis 255 msec
	push r17
	clr r17
	rjmp miL1
milliwait:			;Wartezeit: r17:r16 msec, 1 bis 1000
	push r17
miL1:	push r16
	push r18
	mov r18, r10
milop1:	cp  r18, r10
	breq milop1		;1. Millisekunde abwarten (0 bis 1 ms)
	add r16, r10
	adc r17, r11		;sollzeit = Wartezeit + aktuelle Millisek
	subi r16, low(1001)
	sbci r17, high(1001)	;sollzeit -= 1; sollzeit -= 1000;
	brpl milop2		;wenn sollzeit>=0 -->
	subi r16, low(-1000)	;sonst haben wir 1000 zuviel subtrahiert
	sbci r17, high(-1000)	;also wieder 1000 addieren (soll -= -1000)
milop2:	cp r16, r10
	cpc r17, r11		;auf sollzeit warten
	brne milop2
	cp r16, r10		;falls zwischen 1. u 2. cp ein Interrupt war
	brne milop2
	pop r18
	pop r16
	pop r17
	ret
;-----------------------------------------------------------------------------
