;-----------------------------------------------------------------------------
;* Solarzellen-Nachfuehrung
;* Version: 0.1 	letzte Aenderung: 4.4.2010
;* Prozessor: ATtiny13A, 1.2MHz interner Oszillator
;* Schaltschema:
;*   PB0, PB1 ueber je 2 MOSFET zum Motor,
;*     (2.2k auf G, beide D auf Motor, IRF9620 S nach +5V, MTP3055 S nach GND)
;*   ADC1, ADC3 Analogeingaenge von Phototransistoren,
;*     (PT1331C mit 10k am Collector, sollte bei Sonne weniger 0.1 V ergeben)
;*   ADC3 Analogeingang von Endschaltern,
;*     (oberer Endschalter sollte etwa 0.88 V, unterer etwa 0.45 V ergeben)
;* Autor: Rolf Pfister
;* Copyright: Freeware
;* History:
;* 3.4.2010	Erstellung
;*
;* globale Registerbelegung:
;* r4 = ADC-Kanal
;* r20= PWM-Wert
;* r21= PWM-Verzoegerungs-Zaehler
.equ MAXPWM=16
.equ MINPWM=4

;.equ TEST1=1
;.equ TEST2=1

;-----------------------------------------------------------------------------
.include "tn13def.inc"		; Definitionsdatei fuer den Prozessortyp
; Reset und Interruptvectoren	;VNr. Beschreibung
begin:	rjmp	main		; 1   Power On Reset
	reti			; 2   IRQ0 Handler
	reti			; 3   PCINT0
	rjmp	tc0over		; 4   Timer0 Overflow
	reti			; 5   EEPROM ready
	reti			; 6   Analog Comperator
	reti			; 7   Timer0 CompareA
	reti			; 8   Timer0 CompareB
	reti			; 9   Watchdog Interrupt
	rjmp	adcomplete	; 10  ADC Conversion
;-----------------------------------------------------------------------------
; Variablen im SRAM:
.equ SRAMSTART=0x60
.equ richtung=SRAMSTART		;Motoransteuerung (0b01 oder 0b10, 0b00=aus)
.equ adc1wertL=SRAMSTART+1
.equ adc1wertH=SRAMSTART+2	;Eingelesener Wert von ADC1
.equ adc2wertL=SRAMSTART+3
.equ adc2wertH=SRAMSTART+4	;Eingelesener Wert von ADC1
.equ adc3wertL=SRAMSTART+5
.equ adc3wertH=SRAMSTART+6	;Eingelesener Wert von ADC1

;-----------------------------------------------------------------------------
; Start, Power On, Reset
main:	ldi r16, low(RAMEND)
        out SPL, r16  	     ; Init Stackpointer
.ifdef PRESCALER1
	ldi r16, 0x80
	out CLKPR, r16
	ldi r16, 0
	out CLKPR, r16		;Clk-Prescaler auf 1 setzen --> 9.6MHz
.endif

;; Init-Code
	ldi r16, 0x03		;PB0, PB1 auf Ausgang
        out DDRB, r16		;Data Direction Register B setzen
	ldi r16, 0
	out PORTB, r16		;Motor aus
	ldi r17, 0x02
	out TIMSK0, r17		;Timer/Counter0 Overflow einschalten
	ldi r17, 2		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0B, r17		;Vorteiler auf 8 setzen
;Bei 8MHz wird somit in 8us Schritten gezaehlt (bei 3.6864MHz etwa 8*2.17us)
;Um alle 1000usec einen Interrupt zu erhalten muss der Zaehler jeweils bei
;256-125=131 gestartet werden.
;(bei 3.6864MHz 256-57.6=198.4 oder jeweils 2*57 und 3*58 us)
;.equ	ZAEHLER =131		; 131 bei 8MHz, 198 oder 199 bei 3.6864MHz
;bei 1.2MHz und Vorteiler 8: 6.667us Schritte, 1000/6.667=150, 256-150=106
.equ	ZAEHLER =106		;106 bei 1.2MHz
	ldi r17, ZAEHLER
	out TCNT0, r17
	clr r10
	clr r11			; r11r10 = Millisec
	clr r12			; Sekunden
	clr r13			; Minuten
	clr r14			; Stunden
	clr r15			; Tage, auf 0 setzen

;PWM-Steuerung initialisieren:
	ldi r16, 0
	sts richtung, r16
	ldi r20, 0		;0 oder MINPWM

	sei			;Interrupt erlauben

; AD-Wandler initialisieren:
	ldi r16, 3
	mov r4, r16		;mit ADC3  beginnen
;	ldi r16, 0		;Referenz=VCC
	ldi r16, 0x40		;Referenz= etwa 1.1V
	add r16, r4
	out ADMUX, r16
;Prescaler: 3=8 4=16 5=32 6=64, 1.2MHz/8=150kHz (muss im Bereich 50-200 liegen)
;AD-Wandlung braucht jeweils 13 Takte also etwa 87usec
.equ	PRESCALER=3
;r16 = Prescaler+0b11001000 (ADEN+ADSC+ADIE = Enable,Einzelstart,Interrupt)
;ADIF (0x10) wird automatisch vom AD-Wandler gesetzt wenn Wandlung fertig.
	ldi r16, 0b11001000 + PRESCALER
	out ADCSRA, r16		;Prescaler und Enable
		;ab hier wird also der AD-Wandler dauernd laufen

;-----------------------------------------------------------------------------
mainloop:

.ifdef TEST1
;erster Test: Frequenzmessung an PB0
	cli
t1:	inc r16			;Schlaufe braucht 4 Takte, 8 pro Ganze Periode
	out PORTB, r16		;also 1.2MHz/8 = 150kHz an PB0
	rjmp t1			;oder 9.6MHz/8 = 1.2MHz
.endif

.ifdef TEST2
;Motortest: von einem Anschlag zum andern fahren
	ldi r16, 2		;nach oben fahren
	sts richtung, r16
	out PORTB, r16
wart1:	ldi r18, 220		;auf oberen Endtaster (880mV/4=220) warten
	rcall endtaster
	brne wart1
	ldi r16, 0		;Motor aus
	sts richtung, r16
	out PORTB, r16
	ldi r16, 2
	rcall secwait		;2 Sekunde warten
	ldi r16, 1		;nach unten fahren
	sts richtung, r16
	out PORTB, r16
wart2:	ldi r18, 112		;auf unteren Endtaster (450mV) warten
	rcall endtaster
	brne wart2
	ldi r16, 0		;Motor aus
	sts richtung, r16
	out PORTB, r16
	ldi r16, 2
	rcall secwait		;2 Sekunden warten
	dec r20			;nochmals mit kleinerer Geschwindigkeit
	cpi r20, MAXPWM/2
	brne mainloop
	ldi r20, MAXPWM
	rjmp mainloop

.else
;endgueltige Variante
	lds r16, adc1wertL
	lds r17, adc1wertH	;unterer Lichtwert
	lds r18, adc2wertL
	lds r19, adc2wertH	;oberer Lichtwert
	sub r16, r18
	sbc r17, r19
	breq motstop
	brcs nachunten		;oberer Wert kleiner --> unten heller
nachoben:
	ldi r22, high(500)
	cpi r18, low(500)
	cpc r19, r22		;auf minimale Helligkeit testen
	brcc motstop		;zu dunkel --> keine Nachfuehrung machen
	ldi r18, 220		;auf oberen Endtaster testen
	rcall endtaster
	breq motstop
	rcall pwmsetzen
	ldi r16, 2		;nach oben fahren
	sts richtung, r16
	rjmp mainloop
nachunten:
	rcall negr17r16
	lds r18, adc1wertL
	lds r19, adc1wertH	;unterer Lichtwert
	ldi r22, high(500)
	cpi r18, low(500)
	cpc r19, r22		;auf minimale Helligkeit testen
	brcc motstop		;zu dunkel --> keine Nachfuehrung machen
	ldi r18, 112		;auf unteren Endtaster testen
	rcall endtaster
	breq motstop
	rcall pwmsetzen
	ldi r16, 1		;nach unten fahren
	sts richtung, r16
	rjmp mainloop
motstop:ldi r16, 0		;Motor stoppen
	sts richtung, r16
	out PORTB, r16
	ldi r20, 0		;PWM stoppen
	rjmp mainloop

pwmsetzen:		;PWM abhaengig von r17:r16 setzen, Resultat: r20
	tst r17
	brne mehr
	cpi r16, MINPWM		;r16 kleiner als Minimalwert?
	brcs pwmstop		;ja -->
	cp r20, r16		;r16 groesser oder gleich aktueller PWM-Wert?
	breq return		;gleich -->
	brcs mehr		;groesser -->
	dec r20			;sonst: kleiner
	ret
mehr:	cpi r20, MAXPWM
	breq return
	cp r21, r12		;r21==sec?
	breq return		;Motorgeschwindigkeit nur alle Sek erhoehen
	inc r20
	mov r21, r12
	ret
pwmstop:ldi r20, 0
return:	ret

.endif

.ifdef DIVISIONEN
div18:		;r17:r16 /= 2^r18 ,r18 zerstoert
	asr r17
	ror r16
	dec r18
	brne div18
	ret
.endif

endtaster:			;auf Endtaster warten, Sollwert in r18
	push r16
	push r17
	lds r16, adc3wertL
	lds r17, adc3wertH
	asr r17
	ror r16
	asr r17
	ror r16			;r17:r16 /= 4
	sub r16, r18
	breq ja			;exakt gleich --> fertig: ja
	brcc vergl2		;Ergebnis positiv -->
	neg r16			;
vergl2:	cpi r16, 26		;auf +-100mV gleich?
	brcs ja
	clz			;nein: Zeroflag geloescht
	rjmp pop17ret
ja:	clr r16			;ja: Zeroflag gesetzt
pop17ret:
	pop r17
	pop r16
	ret

vergleich:			;r17:r16 mit r19:r18 vergleichen
	cp r16, r18
	cpc r17, r19

negr17r16:			;r17:r16 = 0 - r17:r16
	com r16
	com r17
	subi r16, -1
	sbci r17, -1
	ret
	
secwait:	; r16 Sekunden lang warten
	push r16
	push r17
	push r18
	mov r18, r16
	add r18, r16
_sw1:	ldi r16, low(500)
	ldi r17, high(500)
	rcall milliwait
	dec r18
	brne _sw1
	pop r18
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; Interrupt-Routine
; Registerbelegungen:
; r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
tc0over:push r16
	push r17
	in r17, SREG	;Statusregister sichern
	push r17
	ldi r17, ZAEHLER
	out TCNT0, r17
	ldi r16, 0xE8
	ldi r17, 0x03	;r17:r16 = 1000 = 0x03E8
	inc r10		;Millisekunden erhoehen
	brne L1
	inc r11
L1:	rcall pwmsteuerung
	cp r10, r16
	brne end1	;fertig wenn msec!=1000
	cp r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
	cp r13, r16
	brne end1	;fertig wenn min!=60
	clr r13		;Minuten auf 0 setzen
	inc r14		;Stunden erhoehen
	ldi r16, 24
	cp r14, r16
	brne end1	;fertig wenn std!=24
	clr r14		;Stunden auf 0 setzen
	inc r15		;Tage erhoehen
end1:
pop17reti:
	pop r17
	out SREG, r17	;Statusregister zurueckholen
	pop r17
	pop r16
	reti

;-----------------------------------------------------------------------------
pwmsteuerung:			; r20=PWM-Wert, 0=aus, 16=Maximum
	push r16
	mov r16, r10
	andi r16, 0x0F		; unterste 4 Bits von millisec verwenden
	cp r16, r20
	brcs pwmein
pwmaus:	clr r16
	rjmp pwmout
pwmein:	lds r16, richtung
pwmout:	out PORTB, r16
	pop r16
	ret

;-----------------------------------------------------------------------------
; Interrupt-Routine fuer AD-Wandler
; Rueckgabewert: adc1wert, adc2wert, adc3wert
adcomplete:
	push r16
	push r17
	in r17, SREG	;Statusregister sichern
	push r17
	push r18
	in r17, ADCL	;L muss zuerst gelesen werden!
	in r16, ADCH
	mov r18, r4
	cpi r18, 3
	breq _ad3
	cpi r18, 2
	breq _ad2
_ad1:	sts adc1wertL, r17 	;Messwert im RAM
	sts adc1wertH, r16	;zwischenspeichern.
	rjmp _ad4
_ad2:	sts adc2wertL, r17 	;Messwert im RAM
	sts adc2wertH, r16	;zwischenspeichern.
	rjmp _ad4
_ad3:	sts adc3wertL, r17 	;Messwert im RAM
	sts adc3wertH, r16	;zwischenspeichern.
_ad4:	dec r4
	brne _ad5
	ldi r16, 3
	mov r4, r16		;ADC3 als Anfangswert
_ad5:	ldi r16, 0x40		;Referenz = etwa 1.1V
	add r16, r4		;ADC-Kanal
	out ADMUX, r16
	sbi ADCSRA, ADSC	;Starte naechste AD-Wandlung
	pop r18
	rjmp pop17reti

;-----------------------------------------------------------------------------
milliwait1:			;Wartezeit: r16 msec, min 1 max 255
	push r17
	clr r17
	rcall milliwait
	pop r17
	ret
milliwait:			;Wartezeit: r17:r16 msec, min 1 max 999
	push r16
	push r17
	push r18
	push r19
	mov r18, r10
	mov r19, r11		;r19r18 = sollzeit = aktuelle Millisec
	add r18, r16
	adc r19, r17		;sollzeit += Wartezeit
	ldi r16, 0xE8		;auf 1000 testen
	cpi r19, 0x03		;oberes Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
	brne subtra		;sollzeit>1000 --> subtrahiere 1000
	cp r18, r16		;unteres Byte testen
	brcs loop1		;sollzeit<1000 --> weiter
subtra:	sub r18, r16
	sbci r19, 0x03		;wenn sollzeit>=1000 dann sollzeit-=1000
loop1:	cp r18, r10
	brne loop1
	cp r19, r11		;auf sollzeit warten
	brne loop1
	pop r19
	pop r18
	pop r17
	pop r16
	ret
